/********************************************************************** 
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/
#ifndef FC__IMPROV_H
#define FC__IMPROV_H

#include "tech.h"
#include "terrain.h"
#include "unit.h"

#ifndef FC__CITY_H
struct city;
#endif

/* Improvement types.
 * (This will later not be an enum.)
 */
enum improvement_type_id {
  B_AIRPORT=0, B_AQUEDUCT, B_BANK, B_BARRACKS, B_BARRACKS2, B_BARRACKS3,
  B_CATHEDRAL, B_CITY, B_COASTAL, B_COLOSSEUM, B_COURTHOUSE,  B_FACTORY,
  B_GRANARY, B_HARBOUR, B_HYDRO, B_LIBRARY, B_MARKETPLACE, B_MASS, B_MFG,
  B_NUCLEAR, B_OFFSHORE, B_PALACE, B_POLICE, B_PORT, B_POWER,
  B_RECYCLING, B_RESEARCH, B_SAM, B_SDI, B_SEWER, B_SOLAR, B_SCOMP,
  B_SMODULE, B_SSTRUCTURAL, B_STOCK, B_SUPERHIGHWAYS, B_SUPERMARKET, B_TEMPLE,
  B_UNIVERSITY,

  B_APOLLO, B_ASMITHS, B_COLLOSSUS, B_COPERNICUS, B_CURE, B_DARWIN, B_EIFFEL,
  B_GREAT, B_WALL, B_HANGING, B_HOOVER, B_ISAAC, B_BACH, B_RICHARDS,
  B_LEONARDO, B_LIGHTHOUSE, B_MAGELLAN, B_MANHATTEN, B_MARCO, B_MICHELANGELO,
  B_ORACLE, B_PYRAMIDS, B_SETI, B_SHAKESPEARE, B_LIBERTY, B_SUNTZU,
  B_UNITED, B_WOMENS,
  B_CAPITAL, B_LAST=100,
/* B_LAST kludged to 100 so that current static arrays should be big enough */
/*
 * FreecivAC Kludge...
 * We _need_ those building ruleset generalisations...
 */
  AC_B_AEROSPACE=0, AC_B_BIOENHANCE, AC_B_BIOLAB, AC_B_PRESERVE,
  AC_B_CRECHE, AC_B_COMMAND, AC_B_ENERGY, AC_B_FUSION, AC_B_GENEJACK,
  AC_B_HAB, AC_B_HAB_DOME, AC_B_HQ, AC_B_NAVAL, AC_B_ORBITMINE,
  AC_B_NODE, AC_B_ORBITDEF, AC_B_ORBITPOW, AC_B_GARDEN,
  AC_B_PERIMETER, AC_B_PRESSURE, AC_B_PSIGATE, AC_B_PUNISHMENT,
  AC_B_CONVERTER, AC_B_QUANTUMLAB,
  AC_B_RECREATION, AC_B_RECYCLE, AC_B_RESHOSPITAL, AC_B_ASSEMBLY,
  AC_B_SKUNK, AC_B_ORBITNUT, AC_B_STOCKPILE, AC_B_TACHYON,
  AC_B_TEMPLE, AC_B_TREEFARM,

  AC_S_ASCETIC, AC_S_MATTER, AC_S_DEFENSE, AC_S_IMMORTALITY,
  AC_S_CLONING, AC_S_NEXUS, AC_S_CYBORG, AC_S_DREAM,
  AC_S_EMPATH, AC_S_GENOME, AC_S_HUNTER, AC_S_REFINERY,
  AC_S_LONGETIVITY, AC_S_MARITIME, AC_S_MERCHANT, AC_S_NANO,
  AC_S_NETWORK, AC_S_NEURAL, AC_S_MUTAGEN, AC_S_DATALINKS,
  AC_S_TRANSIT, AC_S_SELFAWARE, AC_S_SINGULARITY, AC_S_ELEVATOR,
  AC_S_SUPERCOLLIDER, AC_S_TELEPATHIC, AC_S_EVERYTHING,
  AC_S_TRANSLATOR, AC_S_VIRTUAL, AC_S_VOICE, AC_S_WEATHER,
  AC_S_XENOEMPATHY,

  AC_B_LAST
};

typedef int Impr_Type_id;

/* Range of effects.
 * Used in equiv_range and effect.range fields.
 * (These must correspond to effect_range_names[] in city.c.)
 */
enum effect_range_id {
  EFR_NONE,
  EFR_BUILDING,
  EFR_CITY,
  EFR_ISLAND,
  EFR_PLAYER,
  EFR_WORLD,
  EFR_LAST      /* keep this last */
};

typedef enum effect_range_id Eff_Range_id;

/* Type of effects.
 * Used in effect.type field.
 * (These must correspond to effect_type_names[] in city.c.)
 */
enum effect_type_id {
  EFT_ADV_PARASITE,
  EFT_AIRLIFT,
  EFT_ANY_GOVERNMENT,
  EFT_BARB_ATTACK,
  EFT_BARB_DEFEND,
  EFT_CAPITAL_CITY,
  EFT_CAPITAL_EXISTS,
  EFT_ENABLE_NUKE,
  EFT_ENABLE_SPACE,
  EFT_ENEMY_PEACEFUL,
  EFT_FOOD_ADD_TILE,
  EFT_FOOD_BONUS,
  EFT_FOOD_INC_TILE,
  EFT_FOOD_PER_TILE,
  EFT_GIVE_IMM_ADV,
  EFT_GROWTH_FOOD,
  EFT_HAVE_EMBASSIES,
  EFT_IMPROVE_REP,
  EFT_LUXURY_BONUS,
  EFT_LUXURY_PCT,
  EFT_MAKE_CONTENT,
  EFT_MAKE_CONTENT_MIL,
  EFT_MAKE_CONTENT_PCT,
  EFT_MAKE_HAPPY,
  EFT_MAY_DECLARE_WAR,
  EFT_NO_ANARCHY,
  EFT_NO_SINK_DEEP,
  EFT_NUKE_PROOF,
  EFT_POLLU_ADJ,
  EFT_POLLU_ADJ_POP,
  EFT_POLLU_ADJ_PROD,
  EFT_POLLU_DECREASE,
  EFT_PROD_ADD_TILE,
  EFT_PROD_BONUS,
  EFT_PROD_INC_TILE,
  EFT_PROD_PER_TILE,
  EFT_PROD_TO_GOLD,
  EFT_REDUCE_CORRUPT,
  EFT_REDUCE_WASTE,
  EFT_REVEAL_CITIES,
  EFT_REVEAL_MAP,
  EFT_REVOLT_DIST,
  EFT_SCIENCE_BONUS,
  EFT_SCIENCE_PCT,
  EFT_SIZE_UNLIMIT,
  EFT_SLOW_GLOBAL_WARM,
  EFT_SPACE_PART,
  EFT_SPY_RESISTANT,
  EFT_TAX_BONUS,
  EFT_TAX_PCT,
  EFT_TRADE_ADD_TILE,
  EFT_TRADE_BONUS,
  EFT_TRADE_INC_TILE,
  EFT_TRADE_PER_TILE,
  EFT_TRADE_ROUTE_PCT,
  EFT_UNIT_DEFEND,
  EFT_UNIT_MOVE,
  EFT_UNIT_NO_LOSE_POP,
  EFT_UNIT_RECOVER,
  EFT_UNIT_REPAIR,
  EFT_UNIT_VET_COMBAT,
  EFT_UNIT_VETERAN,
  EFT_UPGRADE_UNITS,
  EFT_UPGRADE_ONE_STEP,
  EFT_UPGRADE_ONE_LEAP,
  EFT_UPGRADE_ALL_STEP,
  EFT_UPGRADE_ALL_LEAP,
  EFT_UPKEEP_FREE,
  EFT_LAST      /* keep this last */
};

typedef enum effect_type_id Eff_Type_id;

/* An effect conferred by an improvement.
 */
struct impr_effect {
  Eff_Type_id type;
  Eff_Range_id range;
  int amount;
  Impr_Type_id cond_bldg;               /* B_LAST = unconditional */
  int cond_gov;                         /* game.government_count = unconditional */
  Tech_Type_id cond_adv;                /* A_NONE = unconditional; A_LAST = never */
  Eff_Type_id cond_eff;                 /* EFT_LAST = unconditional */
  Unit_Class_id aff_unit;               /* UCL_LAST = all */
  enum tile_terrain_type aff_terr;      /* T_UNKNOWN = all; T_LAST = none */
  enum tile_special_type aff_spec;      /* S_* bit mask of specials affected */
};

/* Type of improvement.
 * (Read from buildings.ruleset file.)
 */
struct impr_type {
  char name[MAX_LEN_NAME];
  char name_orig[MAX_LEN_NAME];         /* untranslated */
  Tech_Type_id tech_req;                /* A_LAST = never; A_NONE = always */
  Impr_Type_id bldg_req;                /* B_LAST = none required */
  enum tile_terrain_type *terr_gate;    /* list; T_LAST terminated */
  enum tile_special_type *spec_gate;    /* list; S_NO_SPECIAL terminated */
  Eff_Range_id equiv_range;
  Impr_Type_id *equiv_dupl;             /* list; B_LAST terminated */
  Impr_Type_id *equiv_repl;             /* list; B_LAST terminated */
  Tech_Type_id obsolete_by;             /* A_NONE = never obsolete */
  int is_wonder;
  int build_cost;
  int upkeep;
  int sabotage;
  struct impr_effect *effect;           /* list; .type==EFT_LAST terminated */
  int variant;                  /* FIXME: remove when gen-impr obsoletes */
  char *helptext;
};

/* A structure used to keep a track of all effects acting on a city, player,
   etc. as a result of improvement effects (and later, similar effects as
   a result of being in a certain government or nation) */
struct effect {
  struct impr_effect *base;  /* pointer to original impr_effect structure,
                                detailing the actual effect */
  int owner;                 /* id of improvement/government/nation which
                                generated this effect */
  int flags;                 /* See EF_XXX below */
};

/* Bit values for struct effect.flags */
#define EF_ACTIVE   1    /* The effect is active (but dependent on
                            terrain, unit type etc.) */
#define EF_BUILDING 2    /* The effect is conferred by an improvement */
#define EF_GOV      4    /* ditto, for a government */
#define EF_NATION   8    /* ditto, for a nation */


/* Define the effect_list_xxx functions (memory hungry, but better than
   writing our own functions! */
#define SPECLIST_TAG effect
#define SPECLIST_TYPE struct effect
#include "speclist.h"


#define effect_list_iterate(effectlist, peffect) \
    TYPED_LIST_ITERATE(struct effect, effectlist, peffect)
#define effect_list_iterate_end  LIST_ITERATE_END


extern struct impr_type improvement_types[100];

/* improvement effect functions */

Eff_Range_id effect_range_from_str(char *str);
char *effect_range_name(Eff_Range_id id);
Eff_Type_id effect_type_from_str(char *str);
char *effect_type_name(Eff_Type_id id);

int find_improvement_with_effect_type(Eff_Type_id type);

int effect_is_activated(struct effect *neweff,
                        struct effect_list *elist[EFR_LAST],
                        struct city *pcity);
void activate_effect(struct effect *neweff,struct effect_list *elist[EFR_LAST],
                     struct city *pcity);
void add_improvement_effects(struct city *pcity,Impr_Type_id improv);

void are_effects_active(struct impr_effect *effects_to_test,
            struct effect_list *current_effects[EFR_LAST],
            unsigned char *current_improvements,
            struct effect_list active_effects);
void remove_effects(struct impr_effect *effects_to_go,
            struct effect_list *current_effects[EFR_LAST],
            Impr_Type_id impr_type,unsigned char *current_improvements);
void generate_building_effects();
int get_current_effect(Eff_Type_id effect_type,
            struct effect_list *current_effects[EFR_LAST],
            int *amount);

#endif /* FC__IMPROV_H */
